import { getDevOpsCore } from '@thatdevops/core/devopsCore'
import getProjectPkgs from '@thatdevops/core/pkgman/getProjectPkgs'
import Pkg from '@thatdevops/core/pkgman/Pkg'
import echo from '@thatdevops/core/utils/echo'
import spawn from '@thatdevops/core/utils/spawn'
import doNothing from '@thats/utils/misc/doNothing'
import fs from 'fs'
import chunk from 'lodash/chunk'
import { DateTime } from 'luxon'
import path from 'path'
import doFindChangedPackages from './doFindChangedPackages'
import { doPublishPackage } from './doPublishPackage'
import { BumpType, versionBump } from './versionBump'

export async function doCommitPushPublish(props?: { bumpType?: BumpType; commitMessage?: string }) {
	const { parallel = 0 } = getDevOpsCore()
	const { commitMessage = `Publish ${DateTime.now().toISODate()}`, bumpType = BumpType.patch } =
		props ?? {}
	const { publishable } = getDevOpsCore()
	if (!publishable) {
		echo('Publishing not allowed (set the dev ops "publishable" flag')
		return
	}

	// Find packages that have changes since their last publish
	echo('Finding changed packages')
	let changedPkgs = await doFindChangedPackages()
	const effectedPkgs = await doFindEffectedPackages(changedPkgs)

	// Calculate new versions for changed packages
	const newVersions = doCalcNewVersions(effectedPkgs, bumpType)

	// Update any dependencies that have changed
	echo('Updating related dependencies')
	await doUpdateDependencies(effectedPkgs, newVersions)

	echo('Publishing effected packages')
	if (!!parallel) {
		const chunks = chunk(effectedPkgs, parallel)
		for (const chunki of chunks) {
			await Promise.all(chunki.map(c => doPublishPackage(c, newVersions[c.name])))
		}
	} else {
		for (const pkg of effectedPkgs) {
			await doPublishPackage(pkg, newVersions[pkg.name])
		}
	}

	echo(`Committing "${commitMessage}"`)
	await spawn('git', ['add', '-A']).catch(doNothing)
	await spawn('git', ['commit', '-a', '-m', commitMessage]).catch(doNothing)

	echo('Pushing')
	await spawn('git', ['push']).catch(doNothing)
}

export default doCommitPushPublish

function doCalcNewVersions(pkgs: Pkg[], bumpType: BumpType = BumpType.patch) {
	const newVersions = {}
	for (const pkg of pkgs) {
		newVersions[pkg.name] = versionBump(pkg.version, bumpType)
	}
	return newVersions
}

async function doUpdateDependencies(pkgs: Pkg[], newVersions: Record<string, string>) {
	for (const pkg of pkgs) {
		for (const depList of ['dependencies', 'devDependencies']) {
			const deps = pkg.json[depList]
			for (const depName in deps) {
				if (newVersions[depName]) {
					const oldVersion = deps[depName]
					const newVersion = newVersions[depName]
					if (oldVersion !== newVersion) {
						deps[depName] = newVersion
					}
				}
			}
		}
		delete pkg.json.gitHead // Tidying up crap
		fs.writeFileSync(path.join(pkg.location, 'package.json'), JSON.stringify(pkg.json, null, 4))
	}
}

async function doFindEffectedPackages(changedPkgs: Pkg[]): Promise<Pkg[]> {
	const allEffected = []
	let effected = await doFindNewEffectedPackages(changedPkgs)
	allEffected.push(...effected)
	while (effected.length) {
		const newEffected = await doFindNewEffectedPackages(effected)
		effected = newEffected.filter(p => !allEffected.find(p2 => p2.name === p.name))
		allEffected.push(...effected)
	}
	return allEffected
}

async function doFindNewEffectedPackages(changedPkgs: Pkg[]): Promise<Pkg[]> {
	const pkgs = await getProjectPkgs()
	const effected = new Set<Pkg>()
	for (const changedPkg of changedPkgs) {
		effected.add(changedPkg)
		for (const pkg of pkgs) {
			for (const depList of ['dependencies', 'devDependencies']) {
				const deps = pkg.json[depList]
				for (const depName in deps) {
					if (depName === changedPkg.name) {
						effected.add(pkg)
					}
				}
			}
		}
	}
	return Array.from(effected)
}
