import { getDevOpsCore } from '@thatdevops/core/devopsCore'
import { getPkg } from '@thatdevops/core/pkgman/getPkg'
import getProjectPkgMap from '@thatdevops/core/pkgman/getProjectPkgMap'
import getProjectPkgs from '@thatdevops/core/pkgman/getProjectPkgs'
import Pkg from '@thatdevops/core/pkgman/Pkg'
import versionMatch from '@thatdevops/core/pkgman/versionMatch'
import writePkg from '@thatdevops/core/pkgman/writePkg'
import echo from '@thatdevops/core/utils/echo'
import spawnSilently from '@thatdevops/core/utils/spawnSilently'
import Logger from '@thats/utils/log/Logger'
import SimpleMap from '@thats/utils/map/SimpleMap'
import batchPromiseAll from '@thats/utils/misc/batchPromiseAll'
import path from 'path'

export default doResyncPackages
export async function doResyncPackages() {
	const { rootPath, modulePaths = [] } = getDevOpsCore()

	echo('Checking for updated module packages')

	// Get the Package Versions for all the modules
	const allPkgs = await getProjectPkgMap(rootPath)

	await batchPromiseAll(
		modulePaths.map(modulePath => async () => {
			Object.assign(allPkgs, await getProjectPkgMap(modulePath))
		})
	)

	// Process the devops package
	const devopsPkg = getPkg(path.join(rootPath, 'devops'))
	let hasChanged = processPkg(devopsPkg, allPkgs)

	// Process project packages
	const projectPackages = await getProjectPkgs(rootPath)
	projectPackages.forEach(pkg => {
		hasChanged = processPkg(pkg, allPkgs) || hasChanged
	})

	// Yarn install
	if (hasChanged) {
		echo('Yarn installing')
		await spawnSilently('yarn', ['install'], rootPath)
	}
}

function processPkg(pkg: Pkg, allPkgs: SimpleMap<Pkg>) {
	let hasChanged = false

	// Update the Dependency versions
	hasChanged = processDeps(pkg.json.dependencies ?? {}, allPkgs) || hasChanged
	hasChanged = processDeps(pkg.json.devDependencies ?? {}, allPkgs) || hasChanged

	// If the package changed, write it
	if (hasChanged) writePkg(pkg)

	// Done
	return hasChanged
}

function processDeps(deps: SimpleMap<string>, allPkgs: SimpleMap<Pkg>) {
	let hasChanged = false
	for (const dep in deps) {
		if (!!allPkgs[dep]) {
			if (!versionMatch(deps[dep], allPkgs[dep].version)) {
				hasChanged = true
				deps[dep] = allPkgs[dep].version
				Logger.info(` - ${dep} changed to ${allPkgs[dep].version}`)
			}
		}
	}
	return hasChanged
}
