import { getDevOpsCore } from '@thatdevops/core/devopsCore'
import getProjectPkgs from '@thatdevops/core/pkgman/getProjectPkgs'
import Pkg from '@thatdevops/core/pkgman/Pkg'
import echo from '@thatdevops/core/utils/echo'
import { makeRunnable } from '@thatdevops/core/utils/makeRunable'
import fs from 'fs'
import path from 'path'
import { doPublishPackage } from './_/doPublishPackage'
import { BumpType, versionBump } from './_/versionBump'

export const gitCommitBumpPublish = makeRunnable<{
	commitMessage?: string
	deploy?: boolean
	force?: boolean
}>(async o => {
	const { publishable } = getDevOpsCore()
	if (!publishable) {
		echo('Publishing not allowed (set the dev ops "publishable" flag')
		return
	}

	// Find packages that have changes since their last publish
	echo('Finding changed packages')
	// let changedPkgs = await doFindChangedPackages()
	const changedPkgs = (await getProjectPkgs()).filter(
		p => p.relativePath === 'packages/thatweb/actionkeys'
	)
	const effectedPkgs = await doFindEffectedPackages(changedPkgs)

	// Calculate new versions for changed packages
	const newVersions = doCalcNewVersions(effectedPkgs)

	// Update any dependencies that have changed
	echo('Updating related dependencies')
	await doUpdateDependencies(newVersions)

	echo('Publishing changed packages')
	for (const pkg of changedPkgs) {
		await doPublishPackage(pkg, newVersions[pkg.name])
	}

	// echo('Resyncing Package versions')
	// await doResyncPackages()
	//
	// const premessage = 'Prepublish - ' + o.commitMessage + ' [skip-cd]'
	// echo(`Committing "${premessage}"`)
	// await spawn('git', ['add', '-A']).catch(doNothing)
	// await spawn('git', ['commit', '-a', '-m', premessage]).catch(doNothing)
	//
	// echo('Running Yarn Publish')
	// const pkgs = echo(`yarn publish`)
	// console.log('TODO - REWRITE THIS !!! gitCommitBumpPublish')
	// await spawn('yarn', ['publish']).catch(doNothing)
	//
	// const postmessage = 'Postpublish - ' + o.commitMessage + (o.deploy ? '' : ' [skip-cd]')
	// echo(`Committing "${postmessage}"`)
	// await spawn('git', ['commit', '-a', '-m', postmessage]).catch(doNothing)
	//
	// echo(`Pushing to git`)
	// await spawn('git', ['push']).catch(doNothing)
})
export default gitCommitBumpPublish

function doCalcNewVersions(pkgs: Pkg[], bumpType: BumpType = BumpType.patch) {
	const newVersions = {}
	for (const pkg of pkgs) {
		newVersions[pkg.name] = versionBump(pkg.version, bumpType)
	}
	return newVersions
}

async function doUpdateDependencies(newVersions: Record<string, string>) {
	const pkgs = await getProjectPkgs()
	const changed = []
	for (const pkg of pkgs) {
		for (const depList of ['dependencies', 'devDependencies']) {
			const deps = pkg.json[depList]
			for (const depName in deps) {
				if (newVersions[depName]) {
					const oldVersion = deps[depName]
					const newVersion = newVersions[depName]
					if (oldVersion !== newVersion) {
						deps[depName] = newVersion
						changed.push(pkg)
					}
				}
			}
		}
	}
	for (const pkg of changed) {
		fs.writeFileSync(path.join(pkg.location, 'package.json'), JSON.stringify(pkg.json, null, 4))
	}
}

async function doFindEffectedPackages(changedPkgs: Pkg[]): Promise<Pkg[]> {
	const allEffected = []
	let newEffected = await doFindNewEffectedPackages(changedPkgs)
	allEffected.push(...newEffected)
	while (newEffected.length) {
		newEffected = await doFindNewEffectedPackages(newEffected)
		allEffected.push(...newEffected)
	}
	return allEffected
}

async function doFindNewEffectedPackages(changedPkgs: Pkg[]): Promise<Pkg[]> {
	const pkgs = await getProjectPkgs()
	const effected = new Set<Pkg>()
	for (const changedPkg of changedPkgs) {
		effected.add(changedPkg)
		for (const pkg of pkgs) {
			for (const depList of ['dependencies', 'devDependencies']) {
				const deps = pkg.json[depList]
				for (const depName in deps) {
					if (depName === changedPkg.name) {
						effected.add(pkg)
					}
				}
			}
		}
	}
	return Array.from(effected)
}
